#cython: language_level=3
#distutils: language = c++
cimport decoder

cdef class Decoder():
    cdef decoder.Decoder* thisptr

    def __cinit__(self):
        self.thisptr = decoder.decoderCreate()
 
    def __dealloc__(self):
        decoder.decoderRelease( self.thisptr )

    def decode(self, data: bytes):
        cdef void* data_pointer = <void*><char*>data
        cdef size_t data_length = len(data)
        return decoder.decoderDecode(self.thisptr, data_pointer, data_length)

    def get_vertex_count(self):
        return decoder.decoderGetVertexCount(self.thisptr)

    def get_index_count(self):
        return decoder.decoderGetIndexCount(self.thisptr)

    def attribute_is_normalized(self, id: int):
        return decoder.decoderAttributeIsNormalized(self.thisptr, id)

    def read_attribute(self, id: int, componentType: int, dataType: str):
        return decoder.decoderReadAttribute(self.thisptr, id, componentType, dataType.encode("utf-8"))

    def get_attribute_byte_length(self, id: int ):
        return decoder.decoderGetAttributeByteLength(self.thisptr, id)

    def copy_attribute(self, id: int, output: bytes):
        cdef void* output_buffer = <void*><char*>output
        decoder.decoderCopyAttribute(self.thisptr, id, output_buffer)

    def read_indices(self, indexComponentType):
        return decoder.decoderReadIndices(self.thisptr, indexComponentType)

    def get_index_byte_length(self):
        return decoder.decoderGetIndicesByteLength(self.thisptr)

    def copy_indices(self, output: bytes):
        cdef void* output_buffer = <void*><char*>output
        decoder.decoderCopyIndices(self.thisptr, output_buffer)


with section('parse'):
  # Specify structure for custom cmake functions
  additional_commands = {
      'draco_add_emscripten_executable': {
          'kwargs': {
              'NAME': '*',
              'SOURCES': '*',
              'OUTPUT_NAME': '*',
              'DEFINES': '*',
              'INCLUDES': '*',
              'COMPILE_FLAGS': '*',
              'LINK_FLAGS': '*',
              'OBJLIB_DEPS': '*',
              'LIB_DEPS': '*',
              'GLUE_PATH': '*',
              'PRE_LINK_JS_SOURCES': '*',
              'POST_LINK_JS_SOURCES': '*',
              'FEATURES': '*',
          },
          'pargs': 0,
      },
      'draco_add_executable': {
          'kwargs': {
              'NAME': '*',
              'SOURCES': '*',
              'OUTPUT_NAME': '*',
              'TEST': 0,
              'DEFINES': '*',
              'INCLUDES': '*',
              'COMPILE_FLAGS': '*',
              'LINK_FLAGS': '*',
              'OBJLIB_DEPS': '*',
              'LIB_DEPS': '*',
          },
          'pargs': 0,
      },
      'draco_add_library': {
          'kwargs': {
              'NAME': '*',
              'TYPE': '*',
              'SOURCES': '*',
              'TEST': 0,
              'OUTPUT_NAME': '*',
              'DEFINES': '*',
              'INCLUDES': '*',
              'COMPILE_FLAGS': '*',
              'LINK_FLAGS': '*',
              'OBJLIB_DEPS': '*',
              'LIB_DEPS': '*',
              'PUBLIC_INCLUDES': '*',
          },
          'pargs': 0,
      },
      'draco_generate_emscripten_glue': {
          'kwargs': {
              'INPUT_IDL': '*',
              'OUTPUT_PATH': '*',
          },
          'pargs': 0,
      },
      'draco_get_required_emscripten_flags': {
          'kwargs': {
              'FLAG_LIST_VAR_COMPILER': '*',
              'FLAG_LIST_VAR_LINKER': '*',
          },
          'pargs': 0,
      },
      'draco_option': {
          'kwargs': {
              'NAME': '*',
              'HELPSTRING': '*',
              'VALUE': '*',
          },
          'pargs': 0,
      },
      # Rules for built in CMake commands and those from dependencies.
      'list': {
          'kwargs': {
              'APPEND': '*',
              'FILTER': '*',
              'FIND': '*',
              'GET': '*',
              'INSERT': '*',
              'JOIN': '*',
              'LENGTH': '*',
              'POP_BACK': '*',
              'POP_FRONT': '*',
              'PREPEND': '*',
              'REMOVE_DUPLICATES': '*',
              'REMOVE_ITEM': '*',
              'REVERSE': '*',
              'SORT': '*',
              'SUBLIST': '*',
              'TRANSFORM': '*',
          },
      },
      'protobuf_generate': {
        'kwargs': {
            'IMPORT_DIRS': '*',
            'LANGUAGE': '*',
            'OUT_VAR': '*',
            'PROTOC_OUT_DIR': '*',
            'PROTOS': '*',
        },
      },
  }

with section('format'):
  # Formatting options.

  # How wide to allow formatted cmake files
  line_width = 80

  # How many spaces to tab for indent
  tab_size = 2

  # If true, separate flow control names from their parentheses with a space
  separate_ctrl_name_with_space = False

  # If true, separate function names from parentheses with a space
  separate_fn_name_with_space = False

  # If a statement is wrapped to more than one line, than dangle the closing
  # parenthesis on its own line.
  dangle_parens = False

  # Do not sort argument lists.
  enable_sort = False

  # What style line endings to use in the output.
  line_ending = 'unix'

  # Format command names consistently as 'lower' or 'upper' case
  command_case = 'canonical'

  # Format keywords consistently as 'lower' or 'upper' case
  keyword_case = 'upper'



#cython: language_level=3
from libc.stdint cimport uint8_t, uint32_t, uint64_t
from libcpp cimport bool

cdef extern from "encoder.cpp":
    pass

cdef extern from "encoder.h": 

    ctypedef struct Encoder:
        pass

    Encoder* encoderCreate(uint32_t vertexCount)

    void encoderRelease(Encoder *encoder) except *

    void encoderSetCompressionLevel(Encoder* encoder, uint32_t compressionLevel)

    void encoderSetQuantizationBits(Encoder* encoder, uint32_t position, uint32_t normal, uint32_t uv, uint32_t color, uint32_t generic)

    bool encoderEncode(Encoder* encoder, uint8_t preserveTriangleOrder)

    uint64_t encoderGetByteLength(Encoder* encoder)

    void encoderCopy(Encoder*encoder, uint8_t* data)

    void encoderSetIndices(Encoder* encoder, size_t indexComponentType, uint32_t indexCount, void* indices)

    uint32_t encoderSetAttribute(Encoder* encoder, char* attributeName, size_t componentType, char* dataType, void* data)

    uint32_t encoderGetEncodedVertexCount(Encoder* encoder)

    uint32_t encoderGetEncodedIndexCount(Encoder* encoder)


#cython: language_level=3
cimport encoder
from libc.stdint cimport uint8_t

cdef class Encoder():
    cdef encoder.Encoder* thisptr

    def __cinit__(self, vertex_count: int):
        self.thisptr = encoder.encoderCreate(vertex_count)
 
    def __dealloc__(self):
        encoder.encoderRelease( self.thisptr )

    def set_compression_level(self, compressionLevel: int):
        encoder.encoderSetCompressionLevel(self.thisptr, compressionLevel)

    def set_quantization_bits(self, position: int, normal: int, uv: int, color: int, generic: int):
        encoder.encoderSetQuantizationBits(self.thisptr, position, normal, uv, color, generic)

    def encode(self, preserveTriangleOrder: bool = True):
        return encoder.encoderEncode(self.thisptr, preserveTriangleOrder)
    
    def get_byte_length(self):
        return encoder.encoderGetByteLength(self.thisptr)

    def copy(self, output: bytes):
        cdef uint8_t* output_buffer = <uint8_t*>output
        encoder.encoderCopy(self.thisptr, output_buffer)

    def set_indices(self, indexComponentType: int, indexCount: int, indices: bytes):
        cdef void* index_buffer = <void*>indices
        encoder.encoderSetIndices(self.thisptr, indexComponentType, indexCount, index_buffer)

    def set_attribute(self, attributeName: str, componentType: int, dataType: str, data: bytes):
        cdef void* attribute_buffer = <void*><uint8_t*>data
        return encoder.encoderSetAttribute(self.thisptr, attributeName.encode("utf-8"), componentType, dataType.encode("utf-8"), attribute_buffer)

    def get_encoded_vertex_count(self):
        return encoder.encoderGetEncodedVertexCount(self.thisptr)

    def get_encoded_index_count(self):
        return encoder.encoderGetEncodedIndexCount(self.thisptr)



import smtk_draco as draco
import pathlib

test_resource_directory = pathlib.Path(__file__).parent.joinpath("resources/")
testfile = test_resource_directory / "test.drc"

def read_test_file():
    with open(testfile, "rb") as f:
        return f.read()

def get_and_check_attribute(decoder, attribute_id, attribute_name, component_type, data_type, expected_buffer_length, expect_normalized_data ):
    assert decoder.read_attribute(attribute_id, component_type, data_type), f'Expected the attribute {attribute_name} (ID {attribute_id}) to be read'

    # Get the position buffer byte length
    buffer_length = decoder.get_attribute_byte_length(attribute_id)
    assert buffer_length == expected_buffer_length, f"Invalid {attribute_name} buffer size. Expected {expected_buffer_length}, got {buffer_length}"

    # Create a buffer and copy the data
    decoded_buffer_data = bytes(buffer_length)
    decoder.copy_attribute(attribute_id, decoded_buffer_data)

    attribute_is_normalized = decoder.attribute_is_normalized(attribute_id)
    assert expect_normalized_data == attribute_is_normalized, f"Expected attribute_is_normalized to return {expect_normalized_data}, got {not attribute_is_normalized}"

    return decoded_buffer_data

def decode_test_data(test_data: input):
    """ 
    Helper method to decode a Draco data and check the result
    Currently hard-coded for the example file (both raw and encoded).
    """
    decoder = draco.Decoder()

    # Decode the file
    assert decoder.decode(test_data), f"Failed to decode test data"

    # Get vertex count
    vertex_count = decoder.get_vertex_count()
    expected_vertex_count = 6697
    assert vertex_count == expected_vertex_count, f"Invalid vertex count. Expected {expected_vertex_count}, got {vertex_count}"
    
    # Get index count
    index_count = decoder.get_index_count()
    expected_index_count = 12783
    assert index_count == expected_index_count, f"Invalid index count. Expected {expected_index_count}, got {index_count}"
    
    # Read the index data
    assert decoder.read_indices(5123), 'Expected indices to be read'

    # Get the index buffer byte length
    index_buffer_byte_length = decoder.get_index_byte_length()
    expected_index_buffer_byte_length = 25566
    assert index_count == expected_index_count, f"Invalid index buffer size. Expected {expected_index_buffer_byte_length}, got {index_buffer_byte_length}"

    # Create a buffer and copy the data
    index_data = bytes(index_buffer_byte_length)
    decoder.copy_indices(index_data)
    
    # POSITION (attribute 0)
    expected_position_attribute_id = 0
    position_data = get_and_check_attribute(decoder, expected_position_attribute_id, "POSITION", 5126, "VEC3", 80364, False)
    # TODO: validate the data

    # TEXCOORD_0 (attribute 1)
    expected_texcoord_0_attribute_id = 1
    texcoord_0_data = get_and_check_attribute(decoder, expected_texcoord_0_attribute_id, "TEXCOORD_0", 5126, "VEC2", 53576, False)
    # TODO: validate the data

    # NORMAL attribute
    expected_normal_attribute_id = 2
    normal_data = get_and_check_attribute(decoder, expected_normal_attribute_id, "NORMAL", 5126, "VEC3", 80364, False)    
    # TODO: validate the data

    # TANGENT attribute
    expected_tangent_attribute_id = 3
    tangent_data = get_and_check_attribute(decoder, expected_tangent_attribute_id, "TANGENT", 5126, "VEC4", 107152, False)
    # TODO: validate the data

    return vertex_count, index_count, index_data, position_data, texcoord_0_data, normal_data, tangent_data

def encode_test_data( vertex_count, index_count, index_data, position_data, texcoord_0_data, normal_data, tangent_data ): 
    """ 
    Helper method to encode mesh data and check the result
    Currently hard-coded for the example file ()
    """
    encoder = draco.Encoder(vertex_count)

    ## Set the attributes first

    # Set POSITION
    position_attribute_id = encoder.set_attribute("POSITION", 5126, "VEC3", position_data )
    expected_position_attribute_id = 0
    assert position_attribute_id == expected_position_attribute_id, f"Expected POSITION attribute ID to be {expected_position_attribute_id}, got {position_attribute_id}"

    # Set TEXCOORD_0
    texcoord_0_attribute_id = encoder.set_attribute("TEXCOORD_0", 5126, "VEC2", texcoord_0_data )
    expected_texcoord_0_attribute_id = 1
    assert texcoord_0_attribute_id == expected_texcoord_0_attribute_id, f"Expected TEXCOORD_0 attribute ID to be {expected_texcoord_0_attribute_id}, got {texcoord_0_attribute_id}"

    # Set NORMAL
    normal_attribute_id = encoder.set_attribute("NORMAL", 5126, "VEC3", normal_data )
    expected_normal_attribute_id = 2
    assert normal_attribute_id == expected_normal_attribute_id, f"Expected NORMAL attribute ID to be {expected_normal_attribute_id}, got {normal_attribute_id}"
    
    # Set TANGENT
    tangent_attribute_id = encoder.set_attribute("TANGENT", 5126, "VEC4", tangent_data )
    expected_tangent_attribute_id = 3
    assert tangent_attribute_id == expected_tangent_attribute_id, f"Expected TANGENT attribute ID to be {expected_tangent_attribute_id}, got {tangent_attribute_id}"
    
    ## Set the indices
    encoder.set_indices(5123, index_count, index_data)
   
    # Set the compression level
    encoder.set_compression_level(0) # 0 - worst/speediest compression

    # Set the quantization bits
    encoder.set_quantization_bits(position=0, normal=0, uv=0, color=0, generic=0) # 0 = no quantization
    
    ## Encode the the data

    assert encoder.encode(True), "Expected encoding to succeed" # True = Preserve triangle order

    ## Check the encoded data
    encoded_index_count = encoder.get_encoded_index_count()
    assert encoded_index_count == index_count, f"Expected encoded index count {index_count}, got {encoded_index_count}"

    encoded_vertex_count = encoder.get_encoded_vertex_count()
    assert encoded_vertex_count == vertex_count, f"Expected encoded vertex count {vertex_count}, got {encoded_vertex_count}"

    # TODO: decode a model with known encoding settings
    encoded_data_length = encoder.get_byte_length()
    expected_data_length = 347064 # This is bigger than the original. We should really use a sample with known settings
    assert encoded_data_length == expected_data_length, f"Expected encoded data length {expected_data_length}, got {encoded_data_length}"
    
    ## Copy the encoded data
    output_data = bytes(encoded_data_length)
    encoder.copy(output_data)

    del encoder    # Delete the encoder to make sure the __dealloc__ works (just for the purposes of this test)

    return output_data

def test_decoder():
    decode_test_data(read_test_file())

def test_encoder():
    """ 
    Decode a file and then encode it using the encoder.
    Currently, the encoded data is not validated - we basically just check that the methods can be called without a crash.
    """

    # Decode the test file
    (
        initial_vertex_count, 
        initial_index_count, 
        initial_index_data, 
        initial_position_data, 
        initial_texcoord_0_data, 
        initial_normal_data, 
        initial_tangent_data 
    ) = decode_test_data( read_test_file() )
    
    # Encode the data
    encode_test_data(
        initial_vertex_count, 
        initial_index_count, 
        initial_index_data, 
        initial_position_data, 
        initial_texcoord_0_data, 
        initial_normal_data, 
        initial_tangent_data
    )
    # TODO: validate the data
   
def test_decoder_encoder_roundtrip():
    """ 
    Decode a file and then encode it using the encoder, and then decode the result 
    """
    # Decode the test file
    ( 
        initial_vertex_count, 
        initial_index_count, 
        initial_index_data, 
        initial_position_data, 
        initial_texcoord_0_data, 
        initial_normal_data, 
        initial_tangent_data
    ) = decode_test_data( read_test_file() )
    
    # Encode the data
    encoded_data = encode_test_data(
        initial_vertex_count, 
        initial_index_count, 
        initial_index_data, 
        initial_position_data, 
        initial_texcoord_0_data, 
        initial_normal_data, 
        initial_tangent_data
    )

    # Decode the encoded data again and compare the result with the initial decoding result
    (
        decoded_vertex_count, 
        decoded_index_count, 
        decoded_index_data, 
        decoded_position_data, 
        decoded_texcoord_0_data, 
        decoded_normal_data, 
        decoded_tangent_data 
    ) = decode_test_data( encoded_data )

    # Check the decoded vertex and index counts
    assert decoded_vertex_count == initial_vertex_count, f"Expected vertex count {initial_vertex_count}, got {decoded_vertex_count}"
    assert decoded_index_count == initial_index_count, f"Expected vertex count {initial_index_count}, got {decoded_index_count}"
    
    # Check the index data buffer length
    initial_index_byte_count = len(initial_index_data)
    decoded_index_byte_count = len(decoded_index_data)
    assert decoded_index_byte_count == initial_index_byte_count, f"Expected index byte count {initial_index_byte_count}, got {decoded_index_byte_count}"
    # TODO: validate the data
    
    # Check the position data buffer length
    initial_position_byte_count = len(initial_position_data)
    decoded_position_byte_count = len(decoded_position_data)
    assert decoded_position_byte_count == initial_position_byte_count, f"Expected position byte count {initial_position_byte_count}, got {decoded_position_byte_count}"
    # TODO: validate the data

    # Check the texcoord_0 data buffer length
    initial_texcoord_0_byte_count = len(initial_texcoord_0_data)
    decoded_texcoord_0_byte_count = len(decoded_texcoord_0_data)
    assert decoded_texcoord_0_byte_count == initial_texcoord_0_byte_count, f"Expected texcoord_0 byte count {initial_texcoord_0_byte_count}, got {decoded_texcoord_0_byte_count}"
    # TODO: validate the data

    # Check the normal data buffer length
    initial_normal_byte_count = len(initial_normal_data)
    decoded_normal_byte_count = len(decoded_normal_data)
    assert decoded_normal_byte_count == initial_normal_byte_count, f"Expected normal byte count {initial_normal_byte_count}, got {decoded_normal_byte_count}"
    # TODO: validate the data

    # Check the tangent data buffer length
    initial_tangent_byte_count = len(initial_tangent_data)
    decoded_tangent_byte_count = len(decoded_tangent_data)
    assert decoded_tangent_byte_count == initial_tangent_byte_count, f"Expected tangent byte count {initial_tangent_byte_count}, got {decoded_tangent_byte_count}"
    # TODO: validate the data

def test_get_number_of_components():
    assert draco.get_number_of_components("SCALAR") == 1
    assert draco.get_number_of_components("VEC2") == 2
    assert draco.get_number_of_components("VEC3") == 3
    assert draco.get_number_of_components("VEC4") == 4
    assert draco.get_number_of_components("MAT2") == 4
    assert draco.get_number_of_components("MAT3") == 9
    assert draco.get_number_of_components("MAT4") ==  16
    assert draco.get_number_of_components("") == 0 # invalid case (return 0)

def test_get_component_byte_length():
    assert draco.get_component_byte_length(5120) == 1  # Byte
    assert draco.get_component_byte_length(5121) == 1  # Unsigned Byte
    assert draco.get_component_byte_length(5122) == 2  # Short
    assert draco.get_component_byte_length(5123) == 2  # Unsigned Short
    assert draco.get_component_byte_length(5125) == 4  # Unsigned Int
    assert draco.get_component_byte_length(5126) == 4  # Unsigned Float
    assert draco.get_component_byte_length(0) == 0    # invalid case (return 0)

def test_get_attribute_stride():
    # Just some bruteforce testing 😅
    assert draco.get_attribute_stride(5120, "SCALAR") == 1 # Scalar Byte component
    assert draco.get_attribute_stride(5121, "SCALAR") == 1 # Scalar Unsigned Byte component
    assert draco.get_attribute_stride(5122, "SCALAR") == 2 # Scalar Short component
    assert draco.get_attribute_stride(5123, "SCALAR") == 2 # Scalar Unsigned Short component
    assert draco.get_attribute_stride(5125, "SCALAR") == 4 # Scalar Unsigned Int component
    assert draco.get_attribute_stride(5126, "SCALAR") == 4 # Scalar Unsigned Float component

    assert draco.get_attribute_stride(5120, "VEC2") == 2   # Vector with 2 Byte components
    assert draco.get_attribute_stride(5121, "VEC2") == 2   # Vector with 2 Unsigned Byte components
    assert draco.get_attribute_stride(5122, "VEC2") == 4   # Vector with 2 Short components
    assert draco.get_attribute_stride(5123, "VEC2") == 4   # Vector with 2 Unsigned Short components
    assert draco.get_attribute_stride(5125, "VEC2") == 8   # Vector with 2 Unsigned Int components
    assert draco.get_attribute_stride(5126, "VEC2") == 8   # Vector with 2 Unsigned Float components

    assert draco.get_attribute_stride(5120, "VEC3") == 3   # Vector with 3 Byte components
    assert draco.get_attribute_stride(5121, "VEC3") == 3   # Vector with 3 Unsigned Byte components
    assert draco.get_attribute_stride(5122, "VEC3") == 6   # Vector with 3 Short components
    assert draco.get_attribute_stride(5123, "VEC3") == 6   # Vector with 3 Unsigned Short components
    assert draco.get_attribute_stride(5125, "VEC3") == 12  # Vector with 3 Unsigned Int components
    assert draco.get_attribute_stride(5126, "VEC3") == 12  # Vector with 3 Unsigned Float components

    assert draco.get_attribute_stride(5120, "VEC4") == 4   # Vector with 4 Byte components
    assert draco.get_attribute_stride(5121, "VEC4") == 4   # Vector with 4 Unsigned Byte components
    assert draco.get_attribute_stride(5122, "VEC4") == 8   # Vector with 4 Short components
    assert draco.get_attribute_stride(5123, "VEC4") == 8   # Vector with 4 Unsigned Short components
    assert draco.get_attribute_stride(5125, "VEC4") == 16  # Vector with 4 Unsigned Int components
    assert draco.get_attribute_stride(5126, "VEC4") == 16  # Vector with 4 Unsigned Float components

    assert draco.get_attribute_stride(5120, "MAT2") == 4   # Matrix (2x2) with 4 Byte components
    assert draco.get_attribute_stride(5121, "MAT2") == 4   # Matrix (2x2) with 4 Unsigned Byte components
    assert draco.get_attribute_stride(5122, "MAT2") == 8   # Matrix (2x2) with 4 Short components
    assert draco.get_attribute_stride(5123, "MAT2") == 8   # Matrix (2x2) with 4 Unsigned Short components
    assert draco.get_attribute_stride(5125, "MAT2") == 16  # Matrix (2x2) with 4 Unsigned Int components
    assert draco.get_attribute_stride(5126, "MAT2") == 16  # Matrix (2x2) with 4 Unsigned Float components

    assert draco.get_attribute_stride(5120, "MAT3") == 9   # Matrix (3x3) with 9 Byte components
    assert draco.get_attribute_stride(5121, "MAT3") == 9   # Matrix (3x3) with 9 Unsigned Byte components
    assert draco.get_attribute_stride(5122, "MAT3") == 18  # Matrix (3x3) with 9 Short components
    assert draco.get_attribute_stride(5123, "MAT3") == 18  # Matrix (3x3) with 9 Unsigned Short components
    assert draco.get_attribute_stride(5125, "MAT3") == 36  # Matrix (3x3) with 9 Unsigned Int components
    assert draco.get_attribute_stride(5126, "MAT3") == 36  # Matrix (3x3) with 9 Unsigned Float components

    assert draco.get_attribute_stride(5120, "MAT4") == 16  # Matrix (4x4) with 16 Byte components
    assert draco.get_attribute_stride(5121, "MAT4") == 16  # Matrix (4x4) with 16 Unsigned Byte components
    assert draco.get_attribute_stride(5122, "MAT4") == 32  # Matrix (4x4) with 16 Short components
    assert draco.get_attribute_stride(5123, "MAT4") == 32  # Matrix (4x4) with 16 Unsigned Short components
    assert draco.get_attribute_stride(5125, "MAT4") == 64  # Matrix (4x4) with 16 Unsigned Int components
    assert draco.get_attribute_stride(5126, "MAT4") == 64  # Matrix (4x4) with 16 Unsigned Float components

if __name__ == "__main__":
    test_decoder()
    test_encoder()
    test_decoder_encoder_roundtrip()
    test_get_number_of_components()
    test_get_component_byte_length()
    test_get_attribute_stride()
    print("Tests finished")


#cython: language_level=3
#distutils: language = c++

__version__ = "1.0.0"

# Include the implementation files to compile all the code into one extension
include "common.pxi"
include "decoder.pxi"
include "encoder.pxi"


#cython: language_level=3
#distutils: language = c++
cimport common

def get_number_of_components(dataType: str):
    """ 
        Return the number of components in the specified glTF accessor type
        See https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#accessor-data-types
    """
    return common.getNumberOfComponents(dataType.encode("utf-8"))

# TODO: expose ComponentType enum
def get_component_byte_length(componentType):
    """ 
        Return the size in bytes for a value of the specified componentType
        See https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#accessor-data-types
    """
    return common.getComponentByteLength(componentType)

# TODO: expose ComponentType enum
def get_attribute_stride(componentType, dataType: str):
    """ 
        Return the size in bytes of a dataType value consisting of components of componentType.
        
        E.g. the attribute stride of type VEC3 (-> a vector with three components) 
        with unsigned short components (-> 2 bytes) is 6 bytes. For a data buffer containing
        a series of VEC3 values, this simply represents the amount of bytes from the start of
        one VEC3 to the start of the next VEC3.

        See https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#accessor-data-types
    """
    return common.getAttributeStride(componentType, dataType.encode("utf-8"))


#cython: language_level=3
from libc.stdint cimport uint8_t, uint32_t

cdef extern from "common.cpp":
    pass

cdef extern from "common.h": 

    cdef enum ComponentType:
        Byte = 5120,
        UnsignedByte = 5121,
        Short = 5122,
        UnsignedShort = 5123,
        UnsignedInt = 5125,
        Float = 5126,

    size_t getNumberOfComponents(char* dataType)

    size_t getComponentByteLength(size_t componentType)

    size_t getAttributeStride(size_t componentType, char* dataType)



#cython: language_level=3
from libc.stdint cimport uint8_t, uint32_t
from libcpp cimport bool

cdef extern from "decoder.cpp":
    pass

cdef extern from "decoder.h": 

    ctypedef struct Decoder:
        pass

    Decoder* decoderCreate()

    void decoderRelease(Decoder* decoder) except *

    bool decoderDecode(Decoder* decoder, void* data, size_t byteLength)

    uint32_t decoderGetVertexCount(Decoder* decoder)

    uint32_t decoderGetIndexCount(Decoder* decoder)

    bool decoderAttributeIsNormalized(Decoder* decoder, uint32_t id)

    bool decoderReadAttribute(Decoder* decoder, uint32_t id, size_t componentType, char* dataType)

    size_t decoderGetAttributeByteLength(Decoder* decoder, size_t id)

    void decoderCopyAttribute(Decoder* decoder, size_t id, void* output)

    bool decoderReadIndices(Decoder* decoder, size_t indexComponentType)

    size_t decoderGetIndicesByteLength(Decoder* decoder)

    void decoderCopyIndices(Decoder* decoder, void* output)



#!/usr/bin/python3
#
# Copyright 2022 The Draco Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests installations of the Draco library.

Builds the library in shared and static configurations on the current host
system, and then confirms that a simple test application can link in both
configurations.
"""

import argparse
import multiprocessing
import os
import pathlib
import shlex
import shutil
import subprocess
import sys

# CMake executable.
CMAKE = shutil.which('cmake')

# List of generators available in the current CMake executable.
CMAKE_AVAILABLE_GENERATORS = []

# List of variable defs to be passed through to CMake via its -D argument.
CMAKE_DEFINES = []

# CMake builds use the specified generator.
CMAKE_GENERATOR = None

# Enable the transcoder before running tests (sets DRACO_TRANSCODER_SUPPORTED
# and builds transcoder support dependencies).
ENABLE_TRANSCODER = False

# The Draco tree that this script uses.
DRACO_SOURCES_PATH = os.path.abspath(os.path.join('..', '..', '..', '..'))

# Path to this script and the rest of the test project files.
TEST_SOURCES_PATH = os.path.dirname(os.path.abspath(__file__))

# The Draco build directories.
DRACO_SHARED_BUILD_PATH = os.path.join(TEST_SOURCES_PATH, '_draco_build_shared')
DRACO_STATIC_BUILD_PATH = os.path.join(TEST_SOURCES_PATH, '_draco_build_static')

# The Draco install roots.
DRACO_SHARED_INSTALL_PATH = os.path.join(TEST_SOURCES_PATH,
                                         '_draco_install_shared')
DRACO_STATIC_INSTALL_PATH = os.path.join(TEST_SOURCES_PATH,
                                         '_draco_install_static')

DRACO_SHARED_INSTALL_BIN_PATH = os.path.join(DRACO_SHARED_INSTALL_PATH, 'bin')

DRACO_SHARED_INSTALL_LIB_PATH = os.path.join(DRACO_SHARED_INSTALL_PATH, 'lib')

# Argument for -j when using make, or -m when using Visual Studio. Number of
# build jobs.
NUM_PROCESSES = multiprocessing.cpu_count() - 1

# The test project build directories.
TEST_SHARED_BUILD_PATH = os.path.join(TEST_SOURCES_PATH, '_test_build_shared')
TEST_STATIC_BUILD_PATH = os.path.join(TEST_SOURCES_PATH, '_test_build_static')

# The test project install directories.
TEST_SHARED_INSTALL_PATH = os.path.join(TEST_SOURCES_PATH,
                                        '_test_install_shared')
TEST_STATIC_INSTALL_PATH = os.path.join(TEST_SOURCES_PATH,
                                        '_test_install_static')

# Show configuration and build output.
VERBOSE = False


def cmake_get_available_generators():
  """Returns list of generators available in current CMake executable."""
  result = run_process_and_capture_output(f'{CMAKE} --help')

  if result[0] != 0:
    raise Exception(f'cmake --help failed, exit code: {result[0]}\n{result[1]}')

  help_text = result[1].splitlines()
  generators_start_index = help_text.index('Generators') + 3
  generators_text = help_text[generators_start_index::]

  generators = []
  for gen in generators_text:
    gen = gen.split('=')[0].strip().replace('* ', '')

    if gen and gen[0] != '=':
      generators.append(gen)

  return generators


def cmake_get_generator():
  """Returns the CMake generator from CMakeCache.txt in the current dir."""
  cmake_cache_file_path = os.path.join(os.getcwd(), 'CMakeCache.txt')
  cmake_cache_text = ''
  with open(cmake_cache_file_path, 'r') as cmake_cache_file:
    cmake_cache_text = cmake_cache_file.read()

  if not cmake_cache_text:
    raise FileNotFoundError(f'{cmake_cache_file_path} missing or empty.')

  generator = ''
  for line in cmake_cache_text.splitlines():
    if line.startswith('CMAKE_GENERATOR:INTERNAL='):
      generator = line.split('=')[1]

  return generator


def run_process_and_capture_output(cmd, env=None):
  """Runs |cmd| as a child process.

  Returns process exit code and output. Streams process output to stdout when
  VERBOSE is true.

  Args:
    cmd: String containing the command to execute.
    env: Optional dict of environment variables.

  Returns:
    Tuple of exit code and output.
  """
  if not cmd:
    raise ValueError('run_process_and_capture_output requires cmd argument.')

  if os.name == 'posix':
    # On posix systems subprocess.Popen will treat |cmd| as the program name
    # when it is passed as a string. Unconditionally split the command so
    # callers don't need to care about this detail.
    cmd = shlex.split(cmd)

  proc = subprocess.Popen(
      cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)

  if VERBOSE:
    print('COMMAND output:')

  stdout = ''
  for line in iter(proc.stdout.readline, b''):
    decoded_line = line.decode('utf-8')
    if VERBOSE:
      sys.stdout.write(decoded_line)
      sys.stdout.flush()
    stdout += decoded_line

  # Wait for the process to exit so that the exit code is available.
  proc.wait()
  return [proc.returncode, stdout]


def create_output_directories():
  """Creates the build output directores for the test."""
  pathlib.Path(DRACO_SHARED_BUILD_PATH).mkdir(parents=True, exist_ok=True)
  pathlib.Path(DRACO_STATIC_BUILD_PATH).mkdir(parents=True, exist_ok=True)
  pathlib.Path(TEST_SHARED_BUILD_PATH).mkdir(parents=True, exist_ok=True)
  pathlib.Path(TEST_STATIC_BUILD_PATH).mkdir(parents=True, exist_ok=True)


def cleanup():
  """Removes the build output directories from the test."""
  shutil.rmtree(DRACO_SHARED_BUILD_PATH)
  shutil.rmtree(DRACO_STATIC_BUILD_PATH)
  shutil.rmtree(DRACO_SHARED_INSTALL_PATH)
  shutil.rmtree(DRACO_STATIC_INSTALL_PATH)
  shutil.rmtree(TEST_SHARED_BUILD_PATH)
  shutil.rmtree(TEST_STATIC_BUILD_PATH)
  shutil.rmtree(TEST_SHARED_INSTALL_PATH)
  shutil.rmtree(TEST_STATIC_INSTALL_PATH)


def cmake_configure(source_path, cmake_args=None):
  """Configures a CMake build."""
  command = f'{CMAKE} {source_path}'

  if CMAKE_GENERATOR:
    if ' ' in CMAKE_GENERATOR:
      command += f' -G "{CMAKE_GENERATOR}"'
    else:
      command += f' -G {CMAKE_GENERATOR}'

  if cmake_args:
    for arg in cmake_args:
      command += f' {arg}'

  if CMAKE_DEFINES:
    for arg in CMAKE_DEFINES:
      command += f' -D{arg}'

  if VERBOSE:
    print(f'CONFIGURE command:\n{command}')

  result = run_process_and_capture_output(command)

  if result[0] != 0:
    raise Exception(f'CONFIGURE failed!\nexit_code: {result[0]}\n{result[1]}')


def cmake_build(cmake_args=None, build_args=None):
  """Runs a CMake build."""
  command = f'{CMAKE} --build .'

  if cmake_args:
    for arg in cmake_args:
      command += f' {arg}'

  if not build_args:
    build_args = []

  generator = cmake_get_generator()
  if generator.endswith('Makefiles'):
    build_args.append(f'-j {NUM_PROCESSES}')
  elif generator.startswith('Visual'):
    build_args.append(f'-m:{NUM_PROCESSES}')

  if build_args:
    command += ' --'
    for arg in build_args:
      command += f' {arg}'

  if VERBOSE:
    print(f'BUILD command:\n{command}')

  result = run_process_and_capture_output(f'{command}')

  if result[0] != 0:
    raise Exception(f'BUILD failed!\nexit_code: {result[0]}\n{result[1]}')


def run_install_check(install_path):
  """Runs the install_check program."""
  cmd = os.path.join(install_path, 'bin', 'install_check')
  if VERBOSE:
    print(f'RUN command: {cmd}')

  result = run_process_and_capture_output(
      cmd,
      # On Windows, add location of draco.dll into PATH env var
      {'PATH': DRACO_SHARED_INSTALL_BIN_PATH + os.pathsep + os.environ['PATH']},
    )
  if result[0] != 0:
    raise Exception(
        f'install_check run failed!\nexit_code: {result[0]}\n{result[1]}')


def build_and_install_transcoder_dependencies():
  """Builds and installs Draco dependencies for transcoder enabled builds."""
  orig_dir = os.getcwd()

  # The Eigen CMake build in the release Draco has pinned is, to put it mildly,
  # user unfriendly. Instead of wasting time trying to integrate it here, just
  # shutil.copytree() everything in $eigen_submodule_path to
  # $CMAKE_INSTALL_PREFIX/include/Eigen.
  # Eigen claims to be header-only, so this should be adequate for Draco's
  # needs here.
  eigen_submodule_path = os.path.join(
      DRACO_SOURCES_PATH, 'third_party', 'eigen', 'Eigen')

  # "Install" Eigen for the shared install root.
  eigen_install_path = os.path.join(
      DRACO_SHARED_INSTALL_PATH, 'include', 'Eigen')
  shutil.copytree(src=eigen_submodule_path, dst=eigen_install_path)

  # "Install" Eigen for the static install root.
  eigen_install_path = os.path.join(
      DRACO_STATIC_INSTALL_PATH, 'include', 'Eigen')
  shutil.copytree(src=eigen_submodule_path, dst=eigen_install_path)

  # Build and install gulrak/filesystem for shared and static configurations.
  # Note that this is basically running gulrak/filesystem's CMake build as an
  # install script.
  fs_submodule_path = os.path.join(
      DRACO_SOURCES_PATH, 'third_party', 'filesystem')

  # Install gulrak/filesystem in the shared draco install root.
  fs_shared_build = os.path.join(DRACO_SHARED_BUILD_PATH, '_fs')
  pathlib.Path(fs_shared_build).mkdir(parents=True, exist_ok=True)
  os.chdir(fs_shared_build)
  cmake_args = []
  cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={DRACO_SHARED_INSTALL_PATH}')
  cmake_args.append('-DBUILD_SHARED_LIBS=ON')
  cmake_args.append('-DGHC_FILESYSTEM_BUILD_TESTING=OFF')
  cmake_args.append('-DGHC_FILESYSTEM_BUILD_EXAMPLES=OFF')
  cmake_configure(source_path=fs_submodule_path, cmake_args=cmake_args)
  cmake_build(cmake_args=['--target install'])

  # Install gulrak/filesystem in the shared draco install root.
  fs_static_build = os.path.join(DRACO_STATIC_BUILD_PATH, '_fs')
  pathlib.Path(fs_static_build).mkdir(parents=True, exist_ok=True)
  os.chdir(fs_static_build)
  cmake_args = []
  cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={DRACO_SHARED_INSTALL_PATH}')
  cmake_args.append('-DBUILD_SHARED_LIBS=OFF')
  cmake_args.append('-DGHC_FILESYSTEM_BUILD_TESTING=OFF')
  cmake_args.append('-DGHC_FILESYSTEM_BUILD_EXAMPLES=OFF')
  cmake_configure(source_path=fs_submodule_path, cmake_args=cmake_args)
  cmake_build(cmake_args=['--target install'])

  # Build and install TinyGLTF for shared and static configurations.
  # Note, as above, that this is basically running TinyGLTF's CMake build as an
  # install script.
  tinygltf_submodule_path = os.path.join(
      DRACO_SOURCES_PATH, 'third_party', 'tinygltf')

  # Install TinyGLTF in the shared draco install root.
  tinygltf_shared_build = os.path.join(DRACO_SHARED_BUILD_PATH, '_TinyGLTF')
  pathlib.Path(tinygltf_shared_build).mkdir(parents=True, exist_ok=True)
  os.chdir(tinygltf_shared_build)
  cmake_args = []
  cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={DRACO_SHARED_INSTALL_PATH}')
  cmake_args.append('-DTINYGLTF_BUILD_EXAMPLES=OFF')
  cmake_configure(source_path=tinygltf_submodule_path, cmake_args=cmake_args)
  cmake_build(cmake_args=['--target install'])

  # Install TinyGLTF in the static draco install root.
  tinygltf_static_build = os.path.join(DRACO_STATIC_BUILD_PATH, '_TinyGLTF')
  pathlib.Path(tinygltf_static_build).mkdir(parents=True, exist_ok=True)
  os.chdir(tinygltf_static_build)
  cmake_args = []
  cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={DRACO_STATIC_INSTALL_PATH}')
  cmake_args.append('-DTINYGLTF_BUILD_EXAMPLES=OFF')
  cmake_configure(source_path=tinygltf_submodule_path, cmake_args=cmake_args)
  cmake_build(cmake_args=['--target install'])

  os.chdir(orig_dir)


def build_and_install_draco():
  """Builds Draco in shared and static configurations."""
  orig_dir = os.getcwd()

  if ENABLE_TRANSCODER:
    build_and_install_transcoder_dependencies()

  # Build and install Draco in shared library config for the current host
  # machine.
  os.chdir(DRACO_SHARED_BUILD_PATH)
  cmake_args = []
  cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={DRACO_SHARED_INSTALL_PATH}')
  cmake_args.append('-DBUILD_SHARED_LIBS=ON')
  if ENABLE_TRANSCODER:
    cmake_args.append('-DDRACO_TRANSCODER_SUPPORTED=ON')
  cmake_configure(source_path=DRACO_SOURCES_PATH, cmake_args=cmake_args)
  cmake_build(cmake_args=['--target install'])

  # Build and install Draco in the static config for the current host machine.
  os.chdir(DRACO_STATIC_BUILD_PATH)
  cmake_args = []
  cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={DRACO_STATIC_INSTALL_PATH}')
  cmake_args.append('-DBUILD_SHARED_LIBS=OFF')
  if ENABLE_TRANSCODER:
    cmake_args.append('-DDRACO_TRANSCODER_SUPPORTED=ON')
  cmake_configure(source_path=DRACO_SOURCES_PATH, cmake_args=cmake_args)
  cmake_build(cmake_args=['--target install'])

  os.chdir(orig_dir)


def build_test_project():
  """Builds the test application in shared and static configurations."""
  orig_dir = os.getcwd()

  # Configure the test project against draco shared and build it.
  os.chdir(TEST_SHARED_BUILD_PATH)
  cmake_args = []
  cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={TEST_SHARED_INSTALL_PATH}')
  cmake_args.append(f'-DCMAKE_PREFIX_PATH={DRACO_SHARED_INSTALL_PATH}')
  cmake_args.append(f'-DCMAKE_INSTALL_RPATH={DRACO_SHARED_INSTALL_LIB_PATH}')
  cmake_configure(source_path=f'{TEST_SOURCES_PATH}', cmake_args=cmake_args)
  cmake_build(cmake_args=['--target install'])
  run_install_check(TEST_SHARED_INSTALL_PATH)

  # Configure the test project against draco static and build it.
  os.chdir(TEST_STATIC_BUILD_PATH)
  cmake_args = []
  cmake_args.append(f'-DCMAKE_INSTALL_PREFIX={TEST_STATIC_INSTALL_PATH}')
  cmake_args.append(f'-DCMAKE_PREFIX_PATH={DRACO_STATIC_INSTALL_PATH}')
  cmake_configure(source_path=f'{TEST_SOURCES_PATH}', cmake_args=cmake_args)
  cmake_build(cmake_args=['--target install'])
  run_install_check(TEST_STATIC_INSTALL_PATH)

  os.chdir(orig_dir)


def test_draco_install():
  create_output_directories()
  build_and_install_draco()
  build_test_project()
  cleanup()


if __name__ == '__main__':
  CMAKE_AVAILABLE_GENERATORS = cmake_get_available_generators()

  parser = argparse.ArgumentParser()
  parser.add_argument(
      '-G', '--generator', help='CMake builds use the specified generator.')
  parser.add_argument(
      '-D', '--cmake_define',
      action='append',
      help='Passes argument through to CMake as a CMake variable via cmake -D.')
  parser.add_argument(
      '-t', '--with_transcoder',
      action='store_true',
      help='Run tests with Draco transcoder support enabled.')
  parser.add_argument(
      '-v',
      '--verbose',
      action='store_true',
      help='Show configuration and build output.')
  args = parser.parse_args()

  if args.cmake_define:
    CMAKE_DEFINES = args.cmake_define
  if args.generator:
    CMAKE_GENERATOR = args.generator
  if args.verbose:
    VERBOSE = True
  if args.with_transcoder:
    ENABLE_TRANSCODER = True

  if VERBOSE:
    print(f'CMAKE={CMAKE}')
    print(f'CMAKE_DEFINES={CMAKE_DEFINES}')
    print(f'CMAKE_GENERATOR={CMAKE_GENERATOR}')
    print(f'CMAKE_AVAILABLE_GENERATORS={CMAKE_AVAILABLE_GENERATORS}')
    print(f'ENABLE_TRANSCODER={ENABLE_TRANSCODER}')
    print(f'DRACO_SOURCES_PATH={DRACO_SOURCES_PATH}')
    print(f'DRACO_SHARED_BUILD_PATH={DRACO_SHARED_BUILD_PATH}')
    print(f'DRACO_STATIC_BUILD_PATH={DRACO_STATIC_BUILD_PATH}')
    print(f'DRACO_SHARED_INSTALL_PATH={DRACO_SHARED_INSTALL_PATH}')
    print(f'DRACO_STATIC_INSTALL_PATH={DRACO_STATIC_INSTALL_PATH}')
    print(f'NUM_PROCESSES={NUM_PROCESSES}')
    print(f'TEST_SHARED_BUILD_PATH={TEST_SHARED_BUILD_PATH}')
    print(f'TEST_STATIC_BUILD_PATH={TEST_STATIC_BUILD_PATH}')
    print(f'TEST_SOURCES_PATH={TEST_SOURCES_PATH}')
    print(f'VERBOSE={VERBOSE}')

  if CMAKE_GENERATOR and CMAKE_GENERATOR not in CMAKE_AVAILABLE_GENERATORS:
    raise ValueError(f'CMake generator unavailable: {CMAKE_GENERATOR}.')

  test_draco_install()
